import logging
import requests
from requests.exceptions import RequestException
from io import BytesIO
import suds.transport as transport
from atws.wrapper import AutotaskAPIException, ResponseQuery, Wrapper
from atws import wrapper, connection, Query

from django.conf import settings
from djautotask.utils import DjautotaskSettings
from django.core.cache import cache

logger = logging.getLogger(__name__)


class AtwsTransportError(Exception):
    """
    Raise this to indicate exceptions when attempting to get autotask URLs.
    Atws captures requests.RequestException and re-raises them as
    transport.TransportError
    """
    pass


def parse_autotaskprocessexception(e):
    """
    AutotaskProcessException could have any exception wrapped inside of it,
    so handle as an AutotaskAPIException if a response attr is present,
    or handle as we would a regular exception we don't know about.
    """
    response = getattr(e.exception, 'response', None)
    if response:
        msg = str(response.errors)
    else:
        msg = str(e.args[0])
    return msg


def parse_autotaskapiexception(e):
    return ', '.join(get_errors(e.response.errors))


def get_errors(response):
    errors = []
    try:
        for item in response:
            error = item.get('errors')
            message = error[0].get('message')
            errors.append(str(message))
    except (IndexError, TypeError, AttributeError, KeyError):
        # AutotaskAPIException, has given us something unexpected, log
        # it and we can handle this new case.
        logger.exception(str(response))
        errors.append("An unknown Autotask API error.")

    return errors


def init_api_connection(**kwargs):
    client_options = kwargs.setdefault('client_options', {})

    kwargs['apiversion'] = settings.AUTOTASK_CREDENTIALS['api_version']
    kwargs['integrationcode'] = \
        settings.AUTOTASK_CREDENTIALS['integration_code']
    kwargs['username'] = settings.AUTOTASK_CREDENTIALS['username']

    client_options['transport'] = AutotaskRequestsTransport()

    client_options['url'] = get_connection_url(**kwargs)

    return Wrapper(**kwargs)


def get_cached_url(cache_key):
    return cache.get(cache_key)


def get_connection_url(**kwargs):
    cache_key = 'zone_wsdl_url'

    logger.debug('Fetching Autotask wsdl url from cache.')
    wsdl_url_from_cache = get_cached_url(cache_key)
    logger.debug(
        'Cached Autotask {} was: {}'.format(
            cache_key,
            wsdl_url_from_cache
        )
    )

    if not wsdl_url_from_cache:
        try:
            url = connection.get_connection_url(**kwargs)
            cache.set(cache_key, url)
        except transport.TransportError as e:
            raise AtwsTransportError(
                        'Failed to get webservices URL: {}'.format(e)
            )
    else:
        url = wsdl_url_from_cache

    return url


def get_web_url():
    cache_key = 'zone_web_url'
    web_url_from_cache = get_cached_url(cache_key)

    if not web_url_from_cache:
        try:
            url = connection.get_zone_info(
                settings.AUTOTASK_CREDENTIALS['username'],
                settings.AUTOTASK_CREDENTIALS['api_version']
            )['WebUrl']
            cache.set(cache_key, url, timeout=None)
        except transport.TransportError as e:
            raise AtwsTransportError('Failed to get web URL: {}'.format(e))
    else:
        url = web_url_from_cache

    return url


class AutotaskRequestsTransport(transport.Transport):
    # Adapted from atws.connection.RequestsTransport so that we can set
    # our own request settings.

    def __init__(self):
        transport.Transport.__init__(self)

        self.request_settings = DjautotaskSettings().get_settings()
        self.session = requests.Session()
        self.timeout = self.request_settings.get('timeout')
        self.max_attempts = self.request_settings.get('max_attempts')

        self.session.auth = (
            settings.AUTOTASK_CREDENTIALS['username'],
            settings.AUTOTASK_CREDENTIALS['password']
        )
        self.session.mount(
            'https://',
            requests.adapters.HTTPAdapter(
                max_retries=self.request_settings.get('max_attempts'))
        )

    def format_error_message(self, error):
        response = ResponseQuery(wrapper.Wrapper)
        response.add_error(str(error))

        return response

    def open(self, request):
        for attempt in range(1, self.max_attempts + 1):
            try:
                resp = self.session.get(request.url, timeout=self.timeout)
                break
            except RequestException as e:
                if attempt == self.max_attempts:
                    response = self.format_error_message(e)
                    raise AutotaskAPIException(response)
                continue

        return BytesIO(resp.content)

    def send(self, request):
        for attempt in range(1, self.max_attempts + 1):
            try:
                resp = self.session.post(
                    request.url,
                    data=request.message,
                    headers=request.headers,
                    timeout=self.timeout
                )
                break
            except RequestException as e:
                if attempt == self.max_attempts:
                    response = self.format_error_message(e)
                    raise AutotaskAPIException(response)
                continue

        return transport.Reply(
            resp.status_code,
            resp.headers,
            resp.content,
        )


def fetch_object(entity, object_id, at):
    """
    Fetch the give entity from the Autotask.
    """
    query = Query(entity)
    query.WHERE('id', query.Equals, object_id)

    return at.query(query).fetch_one()


def update_object(entity_type, object_id, updated_fields):
    """
    Send updates to Autotask. We need to query for the
    object first, then alter it and execute it.
    https://atws.readthedocs.io/usage.html#querying-for-entities

    This is because we can not create a valid (enough) object to update
    to autotask unless we sync EVERY non-readonly field. If you submit
    the object with no values supplied for the readonly fields,
    Autotask will null them out.
    """
    at = init_api_connection()
    instance = fetch_object(entity_type, object_id, at)

    for key, value in updated_fields.items():
        setattr(instance, key, value)

    return at.update([instance]).fetch_one()


def create_object(entity_type, entity_fields):
    """
    Make a request to Autotask to create the given Autotask entity.
    Returns the created object from the API.
    https://atws.readthedocs.io/usage.html#creating-entities
    """
    at = init_api_connection()
    entity_object = at.new(entity_type)

    for key, value in entity_fields.items():
        setattr(entity_object, key, value)

    return at.create(entity_object).fetch_one()


def update_assigned_resource(at_object, resource, role):
    entity = at_object.__class__.__name__
    query = Query(entity)
    query.WHERE('id', query.Equals, at_object.id)
    at = init_api_connection()

    t = at.query(query).fetch_one()
    t.AssignedResourceID = resource.id if resource else None
    t.AssignedResourceRoleID = role.id if role else None

    # Fetch one executes the update and returns the created object.
    return at.update([t]).fetch_one()



# Generated by Django 2.1.14 on 2020-02-11 10:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0047_auto_20200207_1126'),
    ]

    operations = [
        migrations.AlterField(
            model_name='timeentry',
            name='hours_to_bill',
            field=models.DecimalField(blank=True, decimal_places=4, max_digits=9, null=True),
        ),
        migrations.AlterField(
            model_name='timeentry',
            name='hours_worked',
            field=models.DecimalField(blank=True, decimal_places=4, max_digits=9, null=True),
        ),
        migrations.AlterField(
            model_name='timeentry',
            name='offset_hours',
            field=models.DecimalField(blank=True, decimal_places=4, max_digits=9, null=True),
        ),
    ]



from collections import OrderedDict
from atws.wrapper import AutotaskProcessException, AutotaskAPIException
from xml.sax._exceptions import SAXParseException

from django.core.management.base import BaseCommand, CommandError
from django.utils.translation import ugettext_lazy as _
from djautotask import sync, api

OPTION_NAME = 'autotask_object'
ERROR_MESSAGE_TEMPLATE = 'Failed to sync {}. Autotask API returned an ' \
                             'error(s): {}.'


class Command(BaseCommand):
    help = str(_('Synchronize the specified object with the Autotask API'))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # This can be replaced with a single instantiation of an OrderedDict
        # using kwargs in Python 3.6. But we need Python 3.5 compatibility for
        # now.
        # See https://www.python.org/dev/peps/pep-0468/.
        synchronizers = (
            ('status', sync.StatusSynchronizer, _('Status')),
            ('license_type', sync.LicenseTypeSynchronizer, _('License Type')),
            ('resource', sync.ResourceSynchronizer, _('Resource')),
            ('ticket_secondary_resource',
             sync.TicketSecondaryResourceSynchronizer,
             _('Ticket Secondary Resource')),
            ('priority', sync.PrioritySynchronizer, _('Priority')),
            ('queue', sync.QueueSynchronizer, _('Queue')),
            ('account', sync.AccountSynchronizer, _('Account')),
            ('project_status',
             sync.ProjectStatusSynchronizer, _('Project Status')),
            ('project_type', sync.ProjectTypeSynchronizer, _('Project Type')),
            ('project', sync.ProjectSynchronizer, _('Project')),
            ('phase', sync.PhaseSynchronizer, _('Phase')),
            ('task_secondary_resource',
             sync.TaskSecondaryResourceSynchronizer,
             _('Task Secondary Resource')),
            ('task', sync.TaskSynchronizer, _('Task')),
            ('display_color', sync.DisplayColorSynchronizer,
             _('Display Color')),
            ('ticket_category', sync.TicketCategorySynchronizer,
             _('Ticket Category')),
            ('source', sync.SourceSynchronizer, _('Source')),
            ('issue_type', sync.IssueTypeSynchronizer, _('Issue Type')),
            ('ticket_type', sync.TicketTypeSynchronizer, _('Ticket Type')),
            ('sub_issue_type', sync.SubIssueTypeSynchronizer,
             _('Sub Issue Type')),
            ('ticket', sync.TicketSynchronizer, _('Ticket')),
            ('note_type', sync.NoteTypeSynchronizer, _('Note Type')),
            ('ticket_note', sync.TicketNoteSynchronizer, _('Ticket Note')),
            ('task_note', sync.TaskNoteSynchronizer, _('Task Note')),
            ('task_type_link', sync.TaskTypeLinkSynchronizer,
             _('Task Type Link')),
            ('use_type', sync.UseTypeSynchronizer, _('Use Type')),
            ('allocation_code', sync.AllocationCodeSynchronizer,
             _('Allocation Code')),
            ('role', sync.RoleSynchronizer, _('Role')),
            ('department', sync.DepartmentSynchronizer, _('Department')),
            ('time_entry', sync.TimeEntrySynchronizer, _('Time Entry')),
            (
                'resource_role_department',
                sync.ResourceRoleDepartmentSynchronizer,
                _('Resource Role Department')
            ),
            (
                'resource_service_desk_role',
                sync.ResourceServiceDeskRoleSynchronizer,
                _('Resource Service Desk Role')
            ),
        )
        self.synchronizer_map = OrderedDict()
        for name, synchronizer, obj_name in synchronizers:
            self.synchronizer_map[name] = (synchronizer, obj_name)

    def add_arguments(self, parser):
        parser.add_argument(OPTION_NAME, nargs='?', type=str)
        parser.add_argument('--full',
                            action='store_true',
                            dest='full',
                            default=False)

    def sync_by_class(self, sync_class, obj_name, full_option=False):
        synchronizer = sync_class(full=full_option)

        created_count, updated_count, deleted_count = synchronizer.sync()

        msg = _('{} Sync Summary - Created: {}, Updated: {}')
        fmt_msg = msg.format(obj_name, created_count, updated_count)

        if full_option:
            msg = _('{} Sync Summary - Created: {}, Updated: {}, Deleted: {}')
            fmt_msg = msg.format(obj_name, created_count, updated_count,
                                 deleted_count)

        self.stdout.write(fmt_msg)

    def handle(self, *args, **options):
        sync_classes = []
        autotask_object_arg = options[OPTION_NAME]
        full_option = options.get('full', False)

        if autotask_object_arg:
            object_arg = autotask_object_arg
            sync_tuple = self.synchronizer_map.get(object_arg)

            if sync_tuple:
                sync_classes.append(sync_tuple)
            else:
                msg = _('Invalid AT object {}, '
                        'choose one of the following: \n{}')
                options_txt = ', '.join(self.synchronizer_map.keys())
                msg = msg.format(sync_tuple, options_txt)
                raise CommandError(msg)
        else:
            sync_classes = self.synchronizer_map.values()

        failed_classes = 0
        error_messages = ''

        for sync_class, obj_name in sync_classes:
            error_msg = None
            try:
                self.sync_by_class(sync_class, obj_name,
                                   full_option=full_option)

            except AutotaskProcessException as e:
                error_msg = ERROR_MESSAGE_TEMPLATE.format(
                    obj_name, api.parse_autotaskprocessexception(e))

            except AutotaskAPIException as e:
                error_msg = ERROR_MESSAGE_TEMPLATE.format(
                    obj_name, api.parse_autotaskapiexception(e))

            except SAXParseException as e:
                error_msg = 'Failed to connect to Autotask API. ' \
                      'The error was: {}'.format(e)

            finally:
                if error_msg:
                    self.stderr.write(error_msg)
                    error_messages += '{}\n'.format(error_msg)
                    failed_classes += 1

        if failed_classes > 0:
            msg = '{} class{} failed to sync.\n'.format(
                failed_classes,
                '' if failed_classes == 1 else 'es',
            )
            msg += 'Errors:\n'
            msg += error_messages
            raise CommandError(msg)



# Generated by Django 2.1.11 on 2019-11-27 17:18

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0029_phase_last_activity_date'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='phase',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='djautotask.Phase'),
        ),
    ]



# Generated by Django 2.1.14 on 2020-01-15 16:17

from django.db import migrations, models
import django.db.models.deletion
import django_extensions.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0034_auto_20191210_1518'),
    ]

    operations = [
        migrations.CreateModel(
            name='TimeEntry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('date_worked', models.DateTimeField(blank=True, null=True)),
                ('start_date_time', models.DateTimeField(blank=True, null=True)),
                ('end_date_time', models.DateTimeField(blank=True, null=True)),
                ('summary_notes', models.TextField(blank=True, max_length=8000, null=True)),
                ('internal_notes', models.TextField(blank=True, max_length=8000, null=True)),
                ('non_billable', models.BooleanField(default=False)),
                ('hours_worked', models.DecimalField(blank=True, decimal_places=2, max_digits=9, null=True)),
                ('hours_to_bill', models.DecimalField(blank=True, decimal_places=2, max_digits=9, null=True)),
                ('offset_hours', models.DecimalField(blank=True, decimal_places=2, max_digits=9, null=True)),
                ('resource', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='djautotask.Resource')),
                ('task', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='djautotask.Task')),
                ('ticket', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='djautotask.Ticket')),
            ],
            options={
                'verbose_name_plural': 'Time entries',
            },
        ),
    ]
